package ru.dbotthepony.kstarbound.defs.world.terrain

import com.google.common.collect.ImmutableList
import com.google.common.collect.ImmutableSet
import com.google.gson.JsonElement
import com.google.gson.JsonObject
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kommons.collect.filterNotNull
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.collect.WeightedList
import ru.dbotthepony.kstarbound.defs.AssetReference
import ru.dbotthepony.kstarbound.defs.JsonConfigFunction
import ru.dbotthepony.kstarbound.defs.PerlinNoiseParameters
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import ru.dbotthepony.kstarbound.defs.world.AmbientNoisesDefinition
import ru.dbotthepony.kstarbound.defs.world.BushVariant
import ru.dbotthepony.kstarbound.defs.world.GrassVariant
import ru.dbotthepony.kstarbound.defs.world.Parallax
import ru.dbotthepony.kstarbound.defs.world.SkyColoring
import ru.dbotthepony.kstarbound.defs.world.TreeVariant
import ru.dbotthepony.kstarbound.json.builder.DispatchingAdapter
import ru.dbotthepony.kstarbound.json.builder.IStringSerializable
import ru.dbotthepony.kstarbound.json.builder.JsonFactory
import ru.dbotthepony.kstarbound.json.builder.JsonFlat
import ru.dbotthepony.kstarbound.json.builder.JsonSingleton
import ru.dbotthepony.kstarbound.json.pairListAdapter
import ru.dbotthepony.kstarbound.util.random.AbstractPerlinNoise
import ru.dbotthepony.kstarbound.util.random.nextRange
import ru.dbotthepony.kstarbound.util.random.random
import java.util.random.RandomGenerator
import java.util.stream.IntStream

enum class BiomePlacementDistributionType(
	override val jsonName: String,
	val def: TypeToken<out BiomePlaceablesDefinition.DistributionData>,
	val data: TypeToken<out BiomePlaceables.DistributionData>,
) : IStringSerializable {
	RANDOM("random", TypeToken.get(BiomePlaceablesDefinition.RandomDistribution::class.java), TypeToken.get(BiomePlaceables.RandomDistribution::class.java)),
	PERIODIC("periodic", TypeToken.get(BiomePlaceablesDefinition.PeriodicDistribution::class.java), TypeToken.get(BiomePlaceables.PeriodicDistribution::class.java));

	override fun match(name: String): Boolean {
		return name.lowercase() == jsonName
	}

	companion object {
		val DEFINITION_ADAPTER = DispatchingAdapter("type", { type }, { def }, entries)
		val DATA_ADAPTER = DispatchingAdapter("type", { type }, { data }, entries)
	}
}

enum class BiomePlacementItemType(
	override val jsonName: String,
	val def: TypeToken<out BiomePlaceablesDefinition.DistributionItemData>,
	val data: TypeToken<out BiomePlaceables.Item>,
) : IStringSerializable {
	MICRO_DUNGEON("microdungeon", TypeToken.get(BiomePlaceablesDefinition.MicroDungeon::class.java), TypeToken.get(BiomePlaceables.MicroDungeon::class.java)),
	TREASURE_BOX_SET("treasurebox", TypeToken.get(BiomePlaceablesDefinition.TreasureBox::class.java), TypeToken.get(BiomePlaceables.TreasureBox::class.java)),
	GRASS("grass", TypeToken.get(BiomePlaceablesDefinition.Grass::class.java), TypeToken.get(BiomePlaceables.Grass::class.java)),
	BUSH("bush", TypeToken.get(BiomePlaceablesDefinition.Bush::class.java), TypeToken.get(BiomePlaceables.Bush::class.java)),
	TREE("tree", TypeToken.get(BiomePlaceablesDefinition.Tree::class.java), TypeToken.get(BiomePlaceables.Tree::class.java)),
	OBJECT("object", TypeToken.get(BiomePlaceablesDefinition.Object::class.java), TypeToken.get(BiomePlaceables.Object::class.java)),
	;

	override fun match(name: String): Boolean {
		return name.lowercase() == jsonName
	}

	companion object {
		val DEFINITION_ADAPTER = DispatchingAdapter("type", { type }, { def }, entries)
		val DATA_ADAPTER = DispatchingAdapter("type", { type }, { data }, entries)
	}
}

@JsonFactory
data class BiomePlaceablesDefinition(
	val grassMod: ImmutableList<Registry.Ref<MaterialModifier>> = ImmutableList.of(),
	val grassModDensity: Double = 0.0,
	val ceilingGrassMod: ImmutableList<Registry.Ref<MaterialModifier>> = ImmutableList.of(),
	val ceilingGrassModDensity: Double = 0.0,
	val items: ImmutableList<DistributionItem> = ImmutableList.of(),
) {
	enum class Placement {
		FLOOR, CEILING, BACKGROUND, OCEAN;
	}

	// ----------- ITEMS
	@JsonFactory
	data class DistributionItem(
		val priority: Double = 0.0,
		val variants: Int = 1,
		val mode: Placement = Placement.FLOOR,
		val distribution: AssetReference<DistributionData>,
		@JsonFlat
		val data: DistributionItemData,
	) {
		init {
			checkNotNull(distribution.value) { "Distribution data is missing" }
		}

		fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.DistributionItem {
			return BiomePlaceables.DistributionItem(
				priority = priority,
				variants = variants,
				mode = mode,
				data = distribution.value!!.create(this, biome),
			)
		}
	}

	abstract class DistributionItemData {
		abstract val type: BiomePlacementItemType
		abstract fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item
	}

	@JsonFactory
	data class MicroDungeon(val microdungeons: ImmutableSet<String> = ImmutableSet.of()) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.MICRO_DUNGEON

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			return BiomePlaceables.MicroDungeon(microdungeons)
		}
	}

	@JsonFactory
	data class TreasureBox(val treasureBoxSets: ImmutableSet<String> = ImmutableSet.of()) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.MICRO_DUNGEON

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			return BiomePlaceables.TreasureBox(treasureBoxSets.random(biome.random))
		}
	}

	@JsonFactory
	data class Grass(val grasses: ImmutableSet<Registry.Ref<GrassVariant.Data>> = ImmutableSet.of()) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.GRASS

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			val valid = grasses.stream().map { it.entry }.filterNotNull().toList()

			if (valid.isEmpty()) {
				throw NoSuchElementException("None of grass variants are valid (candidates: $grasses)")
			}

			return BiomePlaceables.Grass(GrassVariant.create(valid.random(biome.random), biome.hueShift))
		}
	}

	@JsonFactory
	data class Tree(
		val treeStemList: ImmutableList<Registry.Ref<TreeVariant.StemData>> = ImmutableList.of(),
		val treeFoliageList: ImmutableList<Registry.Ref<TreeVariant.FoliageData>> = ImmutableList.of(),
		val treeStemHueShiftMax: Double = 0.0,
		val treeFoliageHueShiftMax: Double = 0.0,
		val variantsRange: Vector2i = Vector2i(1, 1),
		val subVariantsRange: Vector2i = Vector2i(2, 2),
		val sameStemHueShift: Boolean = true,
		val sameFoliageHueShift: Boolean = false,
	) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.GRASS

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			val validTreeStems = treeStemList.stream().map { it.entry }.filterNotNull().toList()
			val validFoliage = treeFoliageList.stream().filter { it.key.left().isBlank() || it.entry != null }.toList()

			if (validTreeStems.isEmpty()) {
				throw NoSuchElementException("None of tree stems variants are valid (candidates: $treeStemList)")
			}

			if (validFoliage.isEmpty()) {
				throw NoSuchElementException("None of tree foliage variants are valid (candidates: $treeFoliageList)")
			}

			val pairs = ArrayList<Pair<Registry.Entry<TreeVariant.StemData>, Registry.Entry<TreeVariant.FoliageData>?>>()

			for (stem in validTreeStems) {
				for (foliage in validFoliage) {
					if (foliage.entry == null || stem.value.shape == foliage.value!!.shape) {
						pairs.add(stem to foliage?.entry)
					}
				}
			}

			if (pairs.isEmpty()) {
				throw NoSuchElementException("Out of all possible combinations of tree stems and foliage, none match by shape (stems candidates: $validTreeStems; foliage candidates: $validFoliage)")
			}

			val treeVariants = biome.random.nextRange(variantsRange)

			if (treeVariants <= 0) {
				return BiomePlaceables.Tree(ImmutableList.of())
			}

			val trees = ArrayList<TreeVariant>()

			for (i in 0 until treeVariants) {
				val subTreeVariants = biome.random.nextRange(subVariantsRange)
				if (subTreeVariants <= 0) continue
				val (stem, foliage) = pairs.random(biome.random)

				val treeStemHueShift = treeStemHueShiftMax * biome.random.nextDouble(-1.0, 1.0)
				val treeFoliageHueShift = treeFoliageHueShiftMax * biome.random.nextDouble(-1.0, 1.0)

				if (foliage == null) {
					for (i2 in 0 until subTreeVariants) {
						trees.add(TreeVariant.create(stem, if (sameStemHueShift) treeStemHueShift else treeStemHueShiftMax * biome.random.nextDouble(-1.0, 1.0)))
					}
				} else {
					for (i2 in 0 until subTreeVariants) {
						trees.add(TreeVariant.create(
							stem, if (sameStemHueShift) treeStemHueShift else treeStemHueShiftMax * biome.random.nextDouble(-1.0, 1.0),
							foliage, if (sameFoliageHueShift) treeFoliageHueShift else treeFoliageHueShiftMax * biome.random.nextDouble(-1.0, 1.0)
						))
					}
				}
			}

			return BiomePlaceables.Tree(ImmutableList.copyOf(trees))
		}
	}

	@JsonFactory
	data class BushData(
		val name: Registry.Ref<BushVariant.Data>,
		val baseHueShiftMax: Double = 0.0,
		val modHueShiftMax: Double = 0.0,
	)

	@JsonFactory
	data class Bush(val bushes: ImmutableSet<BushData> = ImmutableSet.of()) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.BUSH

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			val valid = bushes.stream().filter { it.name.entry != null }.toList()

			if (valid.isEmpty()) {
				throw NoSuchElementException("None of bush variants are valid (candidates: $bushes)")
			}

			val (name, baseHueShiftMax, modHueShiftMax) = valid.random(biome.random)
			val data = name.entry!!.value
			val mod = if (data.mods.isNotEmpty()) data.mods.random(biome.random) else ""

			return BiomePlaceables.Bush(BushVariant.create(name.entry!!, biome.random.nextDouble(-1.0, 1.0) * baseHueShiftMax, mod, biome.random.nextDouble(-1.0 * 1.0) * modHueShiftMax))
		}
	}

	@JsonFactory
	data class ObjectPool(val pool: ImmutableList<Pair<Double, String>> = ImmutableList.of(), val parameters: JsonElement = JsonObject())

	@JsonFactory
	data class Object(val objectSets: ImmutableList<ObjectPool>) : DistributionItemData() {
		override val type: BiomePlacementItemType
			get() = BiomePlacementItemType.OBJECT

		override fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables.Item {
			if (objectSets.isEmpty()) {
				return BiomePlaceables.Object(WeightedList())
			}

			val rand = objectSets.random(biome.random)
			return BiomePlaceables.Object(WeightedList(rand.pool.stream().map { it.first to (it.second to rand.parameters) }.collect(ImmutableList.toImmutableList())))
		}
	}

	// --------------- DISTRIBUTION
	abstract class DistributionData {
		abstract val type: BiomePlacementDistributionType
		abstract fun create(self: DistributionItem, biome: BiomeDefinition.CreationParams): BiomePlaceables.DistributionData
	}

	@JsonSingleton
	object RandomDistribution : DistributionData() {
		override val type: BiomePlacementDistributionType
			get() = BiomePlacementDistributionType.RANDOM

		override fun create(
			self: DistributionItem,
			biome: BiomeDefinition.CreationParams
		): BiomePlaceables.DistributionData {
			return BiomePlaceables.RandomDistribution(
				blockSeed = biome.random.nextLong(),
				randomItems = IntStream.range(0, self.variants)
					.mapToObj { self.data.create(biome) }
					.collect(ImmutableList.toImmutableList())
			)
		}
	}

	@JsonFactory
	data class PeriodicDistribution(
		val modulus: Int = 1,
		val octaves: Int = 1,
		val alpha: Double = 2.0,
		val beta: Double = 2.0,
		val modulusVariance: Double = 0.0,
		val densityPeriod: Double = 10.0,
		val densityOffset: Double = 2.0,
		val typePeriod: Double = 10.0,
		val noiseType: PerlinNoiseParameters.Type = PerlinNoiseParameters.Type.PERLIN,
		val noiseScale: Int = PerlinNoiseParameters.DEFAULT_SCALE,
	) : DistributionData() {
		override val type: BiomePlacementDistributionType
			get() = BiomePlacementDistributionType.PERIODIC

		override fun create(
			self: DistributionItem,
			biome: BiomeDefinition.CreationParams
		): BiomePlaceables.DistributionData {
			val modulusOffset = if (modulus == 0) 0 else biome.random.nextInt(-modulus, modulus)

			return BiomePlaceables.PeriodicDistribution(
				modulus = modulus,
				modulusOffset = modulusOffset,

				densityFunction = AbstractPerlinNoise.of(
					PerlinNoiseParameters(
						type = noiseType,
						scale = noiseScale,
						octaves = octaves,
						alpha = alpha,
						beta = beta,

						frequency = 1.0 / densityPeriod,
						amplitude = 1.0,
						bias = densityOffset,
						seed = biome.random.nextLong()
					)
				),

				modulusDistortion = AbstractPerlinNoise.of(
					PerlinNoiseParameters(
						type = noiseType,
						scale = noiseScale,
						octaves = octaves,
						alpha = alpha,
						beta = beta,

						frequency = 1.0 / modulus,
						amplitude = modulusVariance,
						bias = modulusVariance * 2.0,
						seed = biome.random.nextLong()
					)
				),

				weightedItems = IntStream.range(0, self.variants)
					.mapToObj { self.data.create(biome) }
					.map { it to AbstractPerlinNoise.of(
						PerlinNoiseParameters(
							type = noiseType,
							scale = noiseScale,
							octaves = octaves,
							alpha = alpha,
							beta = beta,

							frequency = 1.0 / typePeriod,
							amplitude = 1.0,
							bias = 0.0,
							seed = biome.random.nextLong(),
						)
					) }
					.collect(ImmutableList.toImmutableList())
			)
		}
	}

	fun create(biome: BiomeDefinition.CreationParams): BiomePlaceables {
		return BiomePlaceables(
			grassMod = grassMod.random(biome.random) { null }?.entry,
			ceilingGrassMod = ceilingGrassMod.random(biome.random) { null }?.entry,
			grassModDensity = grassModDensity,
			ceilingGrassModDensity = ceilingGrassModDensity,
			items = items.stream()
				.map { it.create(biome) }
				.collect(ImmutableList.toImmutableList())
		)
	}
}

@JsonFactory
data class BiomeDefinition(
	val airless: Boolean = false,
	val name: String,
	val statusEffects: ImmutableSet<String> = ImmutableSet.of(),
	val weather: ImmutableList<Pair<Double, ImmutableList<AssetReference<WeightedList<String>>>>> = ImmutableList.of(), // binned reference to other assets
	val hueShiftOptions: ImmutableList<Double> = ImmutableList.of(),
	val skyOptions: ImmutableList<SkyColoring> = ImmutableList.of(),
	val description: String = "...",
	val mainBlock: Registry.Ref<TileDefinition>? = null,
	val subBlocks: ImmutableList<Registry.Ref<TileDefinition>>? = null,
	val ores: Registry.Ref<JsonConfigFunction>? = null,
	val musicTrack: AmbientNoisesDefinition? = null,
	val ambientNoises: AmbientNoisesDefinition? = null,
	val surfacePlaceables: BiomePlaceablesDefinition = BiomePlaceablesDefinition(),
	val undergroundPlaceables: BiomePlaceablesDefinition = BiomePlaceablesDefinition(),
	val parallax: AssetReference<Parallax.Data>? = null,
) {
	data class CreationParams(
		val hueShift: Double,
		val random: RandomGenerator
	)

	fun skyColoring(random: RandomGenerator): SkyColoring {
		if (skyOptions.isEmpty())
			return SkyColoring.BLACK

		return skyOptions.random(random)
	}

	fun hueShift(random: RandomGenerator): Double {
		if (hueShiftOptions.isEmpty())
			return 0.0

		return hueShiftOptions.random(random)
	}

	fun create(random: RandomGenerator, verticalMidPoint: Int, threatLevel: Double): Biome {
		val hueShift = hueShift(random)
		val data = CreationParams(hueShift = hueShift, random = random)
		val surfacePlaceables = surfacePlaceables.create(data)
		val undergroundPlaceables = undergroundPlaceables.create(data)

		return Biome(
			baseName = name,
			description = description,
			hueShift = hueShift,

			mainBlock = mainBlock?.entry,
			subBlocks = subBlocks?.stream()?.map { it.entry }?.filterNotNull()?.collect(ImmutableList.toImmutableList()) ?: ImmutableList.of(),

			musicTrack = musicTrack,
			ambientNoises = ambientNoises,
			surfacePlaceables = surfacePlaceables,
			undergroundPlaceables = undergroundPlaceables,

			parallax = parallax?.value?.create(random, verticalMidPoint.toDouble(), hueShift, surfacePlaceables.firstTreeVariant()),

			ores = (ores?.value?.evaluate(threatLevel, oresAdapter)?.map {
				it.stream()
					.filter { it.second > 0.0 }
					.map { Registries.tileModifiers[it.first] to it.second }
					.filter { it.first != null }
					.collect(ImmutableList.toImmutableList())
			}?.orElse(ImmutableList.of()) ?: ImmutableList.of()) as ImmutableList<Pair<Registry.Entry<MaterialModifier>, Double>>
		)
	}

	companion object {
		private val oresAdapter by lazy {
			Starbound.gson.pairListAdapter<String, Double>()
		}
	}
}
