package ru.dbotthepony.kstarbound.world.entities

import com.google.common.collect.ImmutableMap
import com.google.gson.JsonObject
import com.google.gson.TypeAdapter
import com.google.gson.reflect.TypeToken
import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.objects.Object2ObjectOpenHashMap
import ru.dbotthepony.kommons.math.RGBAColor
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.client.StarboundClient
import ru.dbotthepony.kstarbound.client.render.LayeredRenderer
import ru.dbotthepony.kstarbound.client.world.ClientWorld
import ru.dbotthepony.kstarbound.defs.Drawable
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.defs.`object`.ObjectDefinition
import ru.dbotthepony.kstarbound.defs.`object`.ObjectOrientation
import ru.dbotthepony.kstarbound.server.world.ServerWorld
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.set
import ru.dbotthepony.kstarbound.defs.EntityType
import ru.dbotthepony.kstarbound.world.Side
import ru.dbotthepony.kstarbound.world.LightCalculator
import ru.dbotthepony.kstarbound.world.PIXELS_IN_STARBOUND_UNITf
import ru.dbotthepony.kstarbound.world.api.TileColor
import java.io.DataOutputStream
import java.util.HashMap

open class WorldObject(
	val prototype: Registry.Entry<ObjectDefinition>,
) : TileEntity(prototype.file?.computeDirectory() ?: "/") {
	fun deserialize(data: JsonObject) {
		direction = data.get("direction", directions) { Side.LEFT }
		orientationIndex = data.get("orientationIndex", -1)
		interactive = data.get("interactive", false)

		data["uniqueId"]?.let {
			if (!it.isJsonNull)
				uniqueId = it.asString
		}

		for ((k, v) in data.get("parameters") { JsonObject() }.entrySet()) {
			properties[k] = v.deepCopy()
		}
	}

	override val type: EntityType
		get() = EntityType.OBJECT

	override fun writeNetwork(stream: DataOutputStream, isLegacy: Boolean) {
		TODO("Not yet implemented")
	}

	fun serialize(): JsonObject {
		val into = JsonObject()
		into["name"] = prototype.key
		into["tilePosition"] = vectors.toJsonTree(tilePosition)
		into["direction"] = directions.toJsonTree(direction)
		into["orientationIndex"] = orientationIndex
		into["interactive"] = interactive

		if (uniqueId != null) {
			into["uniqueId"] = uniqueId!!
		}

		into["parameters"] = properties.deepCopy()
		return into
	}

	//
	// internal runtime properties
	//
	inline val clientWorld get() = world as ClientWorld
	inline val serverWorld get() = world as ServerWorld
	inline val orientations get() = prototype.value.orientations
	protected val renderParamLocations = HashMap<String, () -> String?>()
	private var frame = 0
		set(value) {
			if (field != value) {
				field = value
				drawablesCache.invalidate()
			}
		}

	private var frameTimer = 0.0
	val flickerPeriod = prototype.value.flickerPeriod?.copy()

	//
	// top level properties
	//
	var uniqueId: String? = null
	var interactive = false
	var direction = Side.LEFT

	var orientationIndex = -1
		set(value) {
			if (field != value) {
				field = value
				invalidate()
			}
		}

	//
	// json driven properties
	//
	var color: TileColor by Property(TileColor.DEFAULT)

	var animationParts: ImmutableMap<String, SpriteReference> by Property()
	var imagePosition: Vector2i by Property(Vector2i.ZERO)
	var animationPosition: Vector2i by Property(Vector2i.ZERO)

	init {
		renderParamLocations["color"] = { color.lowercase }
		renderParamLocations["frame"] = { frame.toString() }
	}

	private val drawablesCache = LazyData {
		orientation?.drawables?.map { it.with(::getRenderParam) } ?: listOf()
	}

	val drawables: List<Drawable> by drawablesCache

	fun getRenderParam(key: String): String? {
		return renderParamLocations[key]?.invoke() ?: "default"
	}

	override fun invalidate(name: String) {
		super.invalidate(name)

		if (name in renderParamLocations) {
			drawablesCache.invalidate()
		}
	}

	override fun invalidate() {
		super.invalidate()
	}

	override fun thinkShared() {
		super.thinkShared()
		flickerPeriod?.update(Starbound.TIMESTEP, world.random)
	}

	override fun thinkRemote() {
		val orientation = orientation

		if (orientation != null) {
			frameTimer = (frameTimer + Starbound.TIMESTEP) % orientation.animationCycle
			frame = (frameTimer / orientation.animationCycle * orientation.frames).toInt()
		}
	}

	val orientation: ObjectOrientation? get() {
		return orientations.getOrNull(orientationIndex)
	}

	override fun defs(): Collection<JsonObject> {
		val get = orientation
		return if (get == null) listOf(prototype.jsonObject) else listOf(get.json, prototype.jsonObject)
	}

	val lightColors: ImmutableMap<String, RGBAColor> by LazyData(listOf("lightColor", "lightColors")) {
		dataValue("lightColor")?.let { ImmutableMap.of("default", colors0.fromJsonTree(it)) }
			?: dataValue("lightColors")?.let { colors1.fromJsonTree(it) }
			?: ImmutableMap.of()
	}

	override fun addLights(lightCalculator: LightCalculator, xOffset: Int, yOffset: Int) {
		var color = lightColors[color.lowercase]

		if (color != null) {
			if (flickerPeriod != null) {
				val sample = flickerPeriod.sinValue().toFloat()
				color *= sample
			}

			lightCalculator.addPointLight(tilePosition.x - xOffset, tilePosition.y - yOffset, color)
		}
	}

	override fun render(client: StarboundClient, layers: LayeredRenderer) {
		val layer = layers.getLayer(orientation?.renderLayer ?: return)

		drawables.forEach {
			val (x, y) = imagePosition
			it.render(client, layer, position.x.toFloat() + x / PIXELS_IN_STARBOUND_UNITf, position.y.toFloat() + y / PIXELS_IN_STARBOUND_UNITf)
		}
	}

	companion object {
		private val colors1 by lazy { Starbound.gson.getAdapter(TypeToken.getParameterized(ImmutableMap::class.java, String::class.java, RGBAColor::class.java)) as TypeAdapter<ImmutableMap<String, RGBAColor>> }
		private val colors0 by lazy { Starbound.gson.getAdapter(RGBAColor::class.java) }
		private val strings by lazy { Starbound.gson.getAdapter(String::class.java) }
		private val directions by lazy { Starbound.gson.getAdapter(Side::class.java) }
		private val vectors by lazy { Starbound.gson.getAdapter(Vector2i::class.java) }

		fun fromJson(content: JsonObject): WorldObject {
			val prototype = Registries.worldObjects[content["name"]?.asString ?: throw IllegalArgumentException("Missing object name")] ?: throw IllegalArgumentException("No such object defined for '${content["name"]}'")
			val pos = content.get("tilePosition", vectors) { throw IllegalArgumentException("No tilePosition was present in saved data") }
			val result = WorldObject(prototype)
			result.tilePosition = pos
			result.deserialize(content)
			return result
		}
	}
}
